# -*- coding: utf-8 -*-
"""
Created on Wed Mar  6 22:01:23 2019

@author: Dingqy
"""

from keras import backend as K
import tensorflow as tf
from rdkit.Chem import rdMolDescriptors
from keras.models import load_model
from rdkit import Chem
import numpy as np
import matplotlib.cm as cm
from rdkit.Chem.Draw import rdMolDraw2D
from IPython.display import SVG

from Model_training_utils import optimize_SVR, optimize_RidgeCV

from TRACE_utils import Parameters

def calculate_gradients_ST(X_train, prev_model):
    '''
    Calculate the gradients for each chemical
    input: X_train --- fingerprint matrix of the chemicals of interest
           prev_model -- trained neural network model
    output: the gradient matrix
    '''
    feed_dict = {}

    with tf.Graph().as_default():
        with tf.Session() as sess:
            K.set_session(sess)
            model = load_model(prev_model)
            feed_dict[model.input.name] = X_train
    
            layer = model.layers[1]
            op_tensor = layer.output
            X = model.layers[0].input
            reconstruct = tf.gradients(op_tensor, X)[0]
    
            out = sess.run(reconstruct, feed_dict = feed_dict)
    return out

def calculate_gradients_baseline(train_dataset, model_type):
    '''
    return the coefients of baseline model after training
    Notice the input dataset must be training dataset!
    '''
    X_new = np.c_[[1]*train_dataset.X.shape[0], train_dataset.X]
    if model_type == 'SVR':
        model = optimize_SVR(X_new, train_dataset.y)
        out = model.coef_
    elif model_type == 'RidgeCV':
        model = optimize_RidgeCV(X_new, train_dataset.y)
        out = model.coef_.flatten()
    else:
        print('model type not supported!')
        return
    return out

def calculate_gradients_TRACE(rown, coln, TRACE_param, k):
    '''
    return the coefient matrix of TRACE model
    '''
    parameter_result = Parameters(rown, coln)
    parameter_result.from_ndarray(TRACE_param, rown, coln)
    
    omega_matrix = parameter_result._omega_matrix
    return omega_matrix[k,:].flatten()

def gradient2atom(smi, gradient, pos_cut = 3, neg_cut = -3, nBits = 2048):
    """
    map the gradient of Morgan fingerprint bit on the molecule
    Input:
        smi - the smiles of the molecule (a string)
        gradient - the 2048 coeffients of the feature
        cutoff - if positive, get the pos where the integrated weight is bigger than the cutoff;
                 if negative, get the pos where the integrated weight is smaller than the cutoff
    Output:
        two list of atom ids (positive and negative)   
    """
    # generate mol 
    mol = Chem.MolFromSmiles(smi)
    # get the bit info of the Morgan fingerprint
    bi = {}
    fp = rdMolDescriptors.GetMorganFingerprintAsBitVect(mol, radius = 2, bitInfo=bi, nBits=nBits)
    onbits = list(fp.GetOnBits())
    # calculate the integrated weight
    atomsToUse = np.zeros((len(mol.GetAtoms()),1))
    for bitId in onbits:
        atomID, radius = bi[bitId][0]
        temp_atomsToUse = []
        if radius > 0:
            env = Chem.FindAtomEnvironmentOfRadiusN(mol, radius, atomID)
            for b in env:
                temp_atomsToUse.append(mol.GetBondWithIdx(b).GetBeginAtomIdx())
                temp_atomsToUse.append(mol.GetBondWithIdx(b).GetEndAtomIdx())
        else:
            temp_atomsToUse.append(atomID)
            env = None
        temp_atomsToUse = list(set(temp_atomsToUse))
        atomsToUse[temp_atomsToUse] += gradient[bitId]
    # get the postively/negatively contributed atom ids
    highlit_pos = []
    highlit_neg = []
    for i in range(len(atomsToUse)):
        if  atomsToUse[i] > pos_cut:
            highlit_pos.append(i)
        elif atomsToUse[i] < neg_cut:
            highlit_neg.append(i)
    return mol, highlit_pos, highlit_neg, atomsToUse

def color_rendering(atomsToUse, cutoff):
    cmap = cm.RdBu_r
    color_dict = {}
    #print(atomsToUse)
    atomsToUse = (atomsToUse.flatten() / cutoff) + 0.5
    for i in range(len(atomsToUse)):
        color_dict[i] = cmap(atomsToUse[i])[0:3]
    return atomsToUse, color_dict

def moltosvg(mol,molSize=(450,200),kekulize=True,drawer=None,**kwargs):
    mc = rdMolDraw2D.PrepareMolForDrawing(mol,kekulize=kekulize)
    if drawer is None:
        drawer = rdMolDraw2D.MolDraw2DSVG(molSize[0],molSize[1])
    drawer.DrawMolecule(mc,**kwargs)
    drawer.FinishDrawing()
    svg = drawer.GetDrawingText()
    return SVG(svg.replace('svg:',''))



# -*- coding: utf-8 -*-
"""
Created on Thu Mar  7 21:13:30 2019

@author: Dingqy
"""

from keras.layers import Dense, Input
from keras.models import Model
from keras.optimizers import Adam
from keras.models import load_model

from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
from sklearn.cluster import MiniBatchKMeans

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.pyplot import savefig
from itertools import cycle
import matplotlib.colors as colors

import numpy as np
import pandas as pd

from rdkit import Chem
from rdkit.Chem import PandasTools
import cairosvg

from bokeh.plotting import figure, output_notebook, show
from bokeh.models import ColumnDataSource

from Model_SAR_viz_utils import calculate_gradients_ST, moltosvg, color_rendering, gradient2atom
from Model_training_utils import prepare_dataset, extract_clean_dataset

#------------------------------------------------
# functions for transfer value calculating (layer1/2/3/4)
def ST_model_layer1(n_features, layer_size, pretrained_params,
            lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0, clipnorm=5):
    training_dat = Input(shape = (n_features,), dtype = 'float32')
    X = Dense(layer_size[0], activation = 'relu')(training_dat)
    model = Model(input = training_dat, output = X)
    optimizer = Adam(lr=lr, beta_1=beta_1, beta_2=beta_2, epsilon=epsilon)
    model.compile(loss = 'mean_squared_error', optimizer=optimizer)
    
    # set parameters
    model.layers[1].set_weights(pretrained_params[0])
    return model

def ST_model_layer2(n_features, layer_size,pretrained_params,
            lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0, clipnorm=5):
    training_dat = Input(shape = (n_features,), dtype = 'float32')
    X = Dense(layer_size[0], activation = 'relu')(training_dat)
    X = Dense(layer_size[1], activation = 'relu')(X)
    model = Model(input = training_dat, output = X)
    optimizer = Adam(lr=lr, beta_1=beta_1, beta_2=beta_2, epsilon=epsilon)
    model.compile(loss = 'mean_squared_error', optimizer=optimizer)
    
    # set parameters
    model.layers[1].set_weights(pretrained_params[0])
    model.layers[2].set_weights(pretrained_params[1])
    return model

def ST_model_layer3(n_features, layer_size, pretrained_params,
            lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0, clipnorm=5):
    training_dat = Input(shape = (n_features,), dtype = 'float32')
    X = Dense(layer_size[0], activation = 'relu')(training_dat)
    X = Dense(layer_size[1], activation = 'relu')(X)
    X = Dense(layer_size[2], activation = 'relu')(X)
    model = Model(input = training_dat, output = X)
    optimizer = Adam(lr=lr, beta_1=beta_1, beta_2=beta_2, epsilon=epsilon)
    model.compile(loss = 'mean_squared_error', optimizer=optimizer)
    
    # set parameters
    model.layers[1].set_weights(pretrained_params[0])
    model.layers[2].set_weights(pretrained_params[1])
    model.layers[3].set_weights(pretrained_params[2])
    return model

def ST_model_layer4(n_features, layer_size, pretrained_params,
            lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0, clipnorm=5):
    training_dat = Input(shape = (n_features,), dtype = 'float32')
    X = Dense(layer_size[0], activation = 'relu')(training_dat)
    X = Dense(layer_size[1], activation = 'relu')(X)
    X = Dense(layer_size[2], activation = 'relu')(X)
    X = Dense(layer_size[3], activation = 'relu')(X)
    model = Model(input = training_dat, output = X)
    optimizer = Adam(lr=lr, beta_1=beta_1, beta_2=beta_2, epsilon=epsilon)
    model.compile(loss = 'mean_squared_error', optimizer=optimizer)
    
    # set parameters
    model.layers[1].set_weights(pretrained_params[0])
    model.layers[2].set_weights(pretrained_params[1])
    model.layers[3].set_weights(pretrained_params[2])
    model.layers[4].set_weights(pretrained_params[3])
    return model

def calculate_transfer_values(prev_model, n_layer):
    model = load_model(prev_model)
    layer1 = model.layers[1]
    param1 = layer1.get_weights()
    layer2 = model.layers[3]
    param2 = layer2.get_weights()
    
    n_features = param1[0].shape[0]
    layer_size = [param1[0].shape[1], param2[0].shape[1]]
    
    if n_layer == 1:
        transfer_model = ST_model_layer1(n_features, layer_size, [param1, param2])
    elif n_layer == 2:
        transfer_model = ST_model_layer2(n_features, layer_size, [param1, param2])
    elif n_layer == 3:
        layer3 = model.layers[5]
        param3 = layer3.get_weights()
        layer_size = [param1[0].shape[1], param2[0].shape[1], param3[0].shape[1]]
        transfer_model = ST_model_layer3(n_features, layer_size, [param1, param2, param3])
    elif n_layer == 4:
        layer3 = model.layers[5]
        param3 = layer3.get_weights()
        layer4 = model.layers[7]
        param4 = layer4.get_weights()
        layer_size = [param1[0].shape[1], param2[0].shape[1], 
                      param3[0].shape[1], param4[0].shape[1]]
        transfer_model = ST_model_layer4(n_features, layer_size, 
                                         [param1, param2, param3, param4])
    return transfer_model

#-----------------------------------------------
def dim_reduce(transfer_model, X):
    transfer_values = transfer_model.predict(X)
    pca = PCA(n_components = 20)
    value_reduced_20d = pca.fit_transform(transfer_values)
    tsne = TSNE(n_components = 2)
    value_reduced = tsne.fit_transform(value_reduced_20d)
    return value_reduced

#------------------------------------------------
def plot_scatter(values, cls, output_name):
    '''
    input: values --- coordinates of each chemical (2 columns)
           cls --- the value for each chemical
           output_name --- name of the output figure

    create a color-map with a different color for each class
    '''
    import matplotlib.cm as cm
    cmap = cm.RdBu_r
    # Get the color for each sample
    normalize = matplotlib.colors.Normalize(vmin=min(cls), vmax=max(cls))
    colors = [cmap(normalize(value)) for value in cls]
    # Extract the x- and y-values
    x = values[:, 0]
    y = values[:, 1]
    # plot
    fig, ax = plt.subplots(figsize=(10,8))
    ax.scatter(x, y, color=colors)
    cax, _ = matplotlib.colorbar.make_axes(ax)
    savefig(output_name)
    plt.show()

#-------------------------------------------------
def cluster_MiniBatch(values, grain_size = 30):
    # clustering using KMeans minibatch algorithm
    n_clusters = int(values.shape[0] / grain_size)
    mbk = MiniBatchKMeans(init = 'k-means++', n_clusters = n_clusters, batch_size=100,
                      n_init = 10, max_no_improvement=10, verbose=0, random_state=0)
    mbk.fit(values)
    return mbk

def plot_clusters(mbk, values):
    mbk.means_labels_unique = np.unique(mbk.labels_)
    colors_ = cycle(colors.cnames.keys())
    n_clusters = mbk.means_labels_unique.size
    fig = plt.figure(figsize=(9,6))
    ax = fig.add_subplot(1, 1, 1)
    for this_centroid, k, col in zip(mbk.cluster_centers_,
                                     range(n_clusters), colors_):
        mask = mbk.labels_ == k
        ax.scatter(values[mask, 0], values[mask, 1], marker='.',
                   c='w', edgecolor=col, alpha=0.5)
        ax.scatter(this_centroid[0], this_centroid[1], marker='+',
                   c='k', s=25)
    ax.set_title("MiniBatchKMeans")
    return

#--------------------------------------------------
def pack_clusters(smiles, labels, molregno, acts, output_prefix, values, cutoff = 4.5):
    '''
    generate sdf files of active clusters for further analysis
    input: smiles --- list of SMILES
           labels --- list of labels (derived from minibatch clustering)
           molregno --- list of chemical ids
           acts --- list of values for chemicals
           output_prefix --- string defining the saving directory and file name prefix
           values --- coordinates of the chemicals (2 columns)
    output: saving sdf files in the directory of destination
    '''
    df = pd.DataFrame({'smi': smiles, 'label': labels, 'molregno': molregno,
                 'Act': acts, 'coord1': values[:,0], 'coord2': values[:,1]})
    df.to_csv(output_prefix + 'store.csv')
    PandasTools.AddMoleculeColumnToFrame(df,'smi','ROMol')
    
    unique_labels = list(set(labels))
    for i in range(len(unique_labels)):
        mask = [item == unique_labels[i] for item in labels]
        y_mean = np.mean(df['Act'].loc[mask])
        if y_mean >= cutoff:
            sub_df = df.loc[mask]
            sdf_file = output_prefix + 'Label_%d.sdf' % unique_labels[i]
            PandasTools.WriteSDF(sub_df, sdf_file, idName='molregno', properties=sub_df.columns)
            print('%d compounds with label %d of mean %.4f: ' % (len(sub_df), unique_labels[i], y_mean))

#-------------------------------------------------
def generate_sdf(df, id_var, fname_right, on_var_right, smi_var_right, 
                 fname_left, on_var_left, smi_var_left, output_sdf_name):
    '''
    merge dataframe containing chemical information and custom dataframe with chemicals of interest, 
              and pack them together in sdf file for further analysis
    input: df --- dataframe containing calculated coordinates for both training dataset and custom dataset
           id_var -- idenifier for df
           fname_right -- name of raw dataset
           on_var_right -- merge identifier for raw dataset
           smi_var_right -- SMILES field for raw dataset
           fname_left -- name of custom dataset
           on_var_left -- merge identifier for custom dataset
           smi_var_left -- SMILES field for custom dataset
           output_sdf_name
    output: sdf_file
    '''

    # prepare right df
    raw_df = pd.read_csv(fname_right)
    info_df = pd.DataFrame({'id': raw_df[on_var_right],
                            'smi': raw_df[smi_var_right]})
    # prepare left df
    left_info_df = pd.read_csv(fname_left)
    left_info_df = pd.DataFrame({'id': left_info_df[on_var_left],
                                 'smi': left_info_df[smi_var_left]})
    info_df = pd.concat([info_df, left_info_df], axis = 0)
    # merge
    new_df = pd.merge(df, info_df, left_on = id_var, right_on = 'id')
    PandasTools.AddMoleculeColumnToFrame(new_df,'smi','ROMol')
    PandasTools.WriteSDF(new_df, output_sdf_name, idName='id', properties=df.columns)
    return new_df

#------------------------------------------------
def sdf2df(sdf_file):
    '''
    convert sdf file to the type of pd.DataFrame
    '''
    suppl = Chem.SDMolSupplier(sdf_file)
    mols = [mol for mol in suppl]
    mol_names = [mol.GetProp('_Name') for mol in mols]
    df = pd.DataFrame({'molregno': mol_names})

    props = list(mols[0].GetPropsAsDict().keys())
    for prop in props:
        values = [mol.GetProp(prop) for mol in mols]
        df[prop] = values
    return df

def df2sdf(df, output_sdf_name, smiles_field, id_field, custom_filter = None):
    '''
    pack pd.DataFrame to sdf_file
    '''
    if custom_filter:
        df = df.loc[custom_filter]
    PandasTools.AddMoleculeColumnToFrame(df,smiles_field,'ROMol')
    PandasTools.WriteSDF(df, output_sdf_name, idName=id_field, properties=df.columns)

    return

def SAR_rendering(X, prev_model, df, id_field, smiles_field, SAR_result_dir, vis_cutoff = 50):
    gradients = calculate_gradients_ST(X, prev_model)
    for k in range(len(df)):
        gradient0 = gradients[k, :]
        smi = df[smiles_field].iloc[k]
        mol, highlit_pos, highlit_neg, atomsToUse = gradient2atom(smi, gradient0)
        atomsToUse, color_dict = color_rendering(atomsToUse, vis_cutoff)
        img = moltosvg(mol, molSize=(450,300), highlightAtoms=[m for m in range(len(atomsToUse))],
            highlightAtomColors = color_dict,highlightBonds=[])
        cairosvg.svg2png(bytestring=img.data, write_to = SAR_result_dir + '%s_img.png' % str(df[id_field].iloc[k]))
    png_file_names = [SAR_result_dir + str(item) + '_img.png' for item in df[id_field].tolist()]
    df['imgs'] = png_file_names
    return df

def interactive_plot(plot_df, x_column, y_column, color_column, id_field, value_field, label_field):
    # color rendering
    from bokeh.palettes import RdBu11
    COLORS = RdBu11
    N_COLORS = len(COLORS)
    groups = pd.qcut(plot_df[color_column].tolist(), N_COLORS, duplicates='drop')
    c = [COLORS[xx] for xx in groups.codes]
    plot_df['c'] = c

    # prepare bokeh dataset
    cds_df = ColumnDataSource(plot_df)

    # set hover display format
    TOOLTIPS = """
        <div>
            <div>
                <img
                    src="@imgs" height="300" alt="@imgs" width="240"
                    style="float: left; margin: 0px 15px 15px 0px;"
                    border="2"
                ></img>
            </div>
            <div>
                <span style="font-size: 17px; font-weight: bold;">@%s</span>
                <span style="font-size: 15px; color: #966;">[$index]</span>
            </div>
            <div>
                <span style="font-size: 10px;">Activity:</span>
                <span>@%s</span>
            </div>
            <div>
                <span style="font-size: 10px;">Cluster label:</span>
                <span>@%s</span>
            </div>
            <div>
                <span style="font-size: 15px;">Location</span>
                <span style="font-size: 10px; color: #696;">($x, $y)</span>
            </div>
        </div>
    """ % (id_field, value_field, label_field)
    p = figure(plot_width = 800, plot_height = 800, toolbar_location = 'below',
               tools = 'pan,box_zoom,reset,lasso_select,save,hover', tooltips=TOOLTIPS)
    p.circle(source = cds_df, x = x_column, y = y_column, color='c', size=9)
    return p

#================================================
def landscape_building(task_name, db_name, log_path, FP_type,
                       prev_model, n_layer, 
                       SAR_result_dir, output_sdf_name,
                       interactive_plot = True, pack_sdf = True, vis_cutoff = 50,
                       smiles_field = 'salt_removed_smi', id_field = 'molregno'):
    '''
    generate chemical landscape for a specific task
    '''
    # step1: prepare dataset for the landscape
    print('==== preparing dataset ... ====')
    dataset_file = '%s/temp.csv' % (log_path)
    dataset, df = prepare_dataset(db_name, [task_name], dataset_file, FP_type)

    # step2: calculate transfer value
    print('==== calculating transfer values ... ====')
    transfer_model = calculate_transfer_values(prev_model, n_layer)
    value_reduced = dim_reduce(transfer_model, dataset.X)
    df['coord1'] = value_reduced[:,0]
    df['coord2'] = value_reduced[:,1]

    # step3: MiniBatch clustering
    mbk = cluster_MiniBatch(value_reduced)
    df['Label'] = mbk.labels_

    # step4: SAR rendering for all compounds
    print('==== rendering SAR for chemicals on the landscape ... ====')
    df = SAR_rendering(dataset.X, prev_model, df, id_field, smiles_field, SAR_result_dir, vis_cutoff)

    # step5: pack sdf
    if pack_sdf:
        print('==== packing sdf file ... ====')
        df2sdf(df, output_sdf_name, smiles_field, id_field)

    # step6: build interactive plot using Bokeh
    plot_df = df.drop(['ROMol'], axis = 1)
    if interactive_plot:
        print('==== generate interactive plot ... ====')
        p = interactive_plot(plot_df, 'coord1', 'coord2', task_name,
                             id_field, task_name, 'Label')
        show(p)
    return


def landscape_positioning(custom_file, custom_smi_field, custom_id_field, custom_task_field,
                          task_name, db_name, FP_type, log_path,
                          prev_model, n_layer, custom_SAR_result_dir, custom_sdf_name,
                          interactive_plot = True, pack_sdf = True, vis_cutoff = 50,
                          smiles_field = 'salt_removed_smi', id_field = 'molregno'):
    # step1: prepare dataset for the landscape
    print('==== preparing dataset ... ====')
    dataset_file = '%s/temp.csv' % (log_path)
    dataset, df = prepare_dataset(db_name, [task_name], dataset_file, FP_type)

    custom_dataset, custom_df = prepare_dataset(custom_file, [custom_task_field], custom_file, FP_type, 
                            smiles_field = custom_smi_field, id_field = custom_id_field)

    # step2: calculate transfer value
    print('==== calculating transfer values ... ====')
    transfer_model = calculate_transfer_values(prev_model, n_layer)
    X_new = np.concatenate([custom_dataset.X, dataset.X])
    value_reduced = dim_reduce(transfer_model, X_new)
    
    custom_df['coord1'] = value_reduced[0:len(custom_df),0]
    custom_df['coord2'] = value_reduced[0:len(custom_df),1]
    df['coord1'] = value_reduced[len(custom_df):value_reduced.shape[0],0]
    df['coord2'] = value_reduced[len(custom_df):value_reduced.shape[0]:,1]

    # step3: MiniBatch clustering
    mbk = cluster_MiniBatch(value_reduced)
    custom_df['Label'] = mbk.labels_[0:len(custom_df)]
    df['Label'] = mbk.labels_[len(custom_df):value_reduced.shape[0]]

    # step4: SAR rendering for all compounds
    print('==== rendering SAR for chemicals on the landscape ... ====')
    custom_df = SAR_rendering(custom_dataset.X, prev_model, custom_df, custom_id_field, custom_smi_field, 
                              custom_SAR_result_dir, vis_cutoff)

    # step5: pack sdf
    if pack_sdf:
        print('==== packing sdf file ... ====')
        plot_df = pd.DataFrame({'coord1': value_reduced[:,1], 'coord2': value_reduced[:,1]})
        plot_df[id_field] = custom_df[custom_id_field].tolist() + df[id_field].tolist()
        plot_df[task_name] = custom_df[custom_task_field].tolist() + df[task_name].tolist()
        plot_df['Label'] = mbk.labels_
        plot_df[smi_field] = custom_df[custom_smi_field].tolist() + df[smiles_field].tolist()
        df2sdf(plot_df, custom_sdf_name, smi_field, id_field)

    # step6: build interactive plot using Bokeh
    plot_df = plot_df.drop(['ROMol'], axis = 1)
    if interactive_plot:
        print('==== generate interactive plot ... ====')
        plot_df['group'] = [1] * len(custom_df) + [0] * len(df)
        p = interactive_plot(plot_df, 'coord1', 'coord2', 'group',
                             id_field, task_name, 'Label')
        show(p)
    return custom_df






# -*- coding: utf-8 -*-
"""
Created on Wed Mar  6 17:07:01 2019

@author: Dingqy
"""
import os

import numpy as np
import deepchem as dc
import pandas as pd

from rdkit import Chem
from rdkit.Chem import AllChem, MACCSkeys

from keras.layers import Dense, Input
from keras.layers.core import Dropout
from keras.models import Model
from keras.optimizers import Adam
from keras.callbacks import ModelCheckpoint

from sklearn import linear_model
from sklearn.svm import LinearSVR

from scipy.stats import pearsonr
from sklearn.metrics import mean_squared_error
from math import sqrt
from sklearn import preprocessing


#---------------------------------------------
# baseline model training
def optimize_SVR(X_train, Y_train):
    # input: X_train --- each row a sample; each col a feature
    #        Y_train 
    # output: a predictive model of support vector regression
    clf = LinearSVR(C=1.0, epsilon=0.2)
    try:
        clf.fit(X_train, Y_train)
    except Exception as e:
        print(e)
    return clf

def optimize_RidgeCV(X_train, Y_train):
    '''
    input: X_train --- each row a sample; each col a feature
           Y_train 
    output: a predictive model of ridge regression with selected alpha by cross-validation
    '''
    alphas = np.logspace(start = -1, stop = 2, num = 20)
    reg = linear_model.RidgeCV(alphas)
    try:
        reg.fit(X_train, Y_train)
    except Exception as e:
        print(e)
    return reg

#---------------------------------------------
def prepare_dataset(fname, task, dataset_file, FP_type, 
                    model_flag = 'ST', add_features = None,
                    smiles_field = 'salt_removed_smi', id_field = 'molregno'):
    '''
    input: fname --- name of the file of raw data containing chemicals and the value for each assay;
           task --- list of task names (supporting 1 or more, but must be a list)
           dataset_file --- name of the temperary file saving intermediate dataset containing only the data for a specific training task
           FP_type --- name of the fingprint type (one of 'Circular_2018', 'Circular_1024', 'Morgan', 'MACCS', 'RDKit_FP')
           model_flag --- type of the model (ST: single task, len(task)==1; MT: multitask, len(task)>=1)
           add_features --- must be None for ST; list of properties
           smiles_field --- the name of the field in fname defining the SMILES for chemicals
           id_field --- the name of the field in fname defining the identifier for chemicals
    output: dataset build by deepchem
    '''
    MT_df = pd.read_csv(fname)
    if model_flag == 'ST':
        df = extract_clean_dataset(task, MT_df, smiles_field = smiles_field, id_field = id_field)
    elif model_flag == 'MT':
        df = extract_clean_dataset(task, MT_df, add_features = ['MW','logP','TPSA','BertzCT'], smiles_field = smiles_field, id_field = id_field)
        task = task + add_features
    
    if FP_type == 'Circular_2048':
        df.to_csv(dataset_file)
        featurizer = dc.feat.CircularFingerprint(size=2048)
        loader = dc.data.CSVLoader(id_field=id_field, 
                                   smiles_field=smiles_field, 
                                   tasks = task,
                                   featurizer=featurizer)
        dataset = loader.featurize(dataset_file)
        
    elif FP_type == 'Circular_1024':
        df.to_csv(dataset_file)
        featurizer = dc.feat.CircularFingerprint(size=1024)
        loader = dc.data.CSVLoader(id_field=id_field, 
                                   smiles_field=smiles_field, 
                                   tasks = task,
                                   featurizer=featurizer)
        dataset = loader.featurize(dataset_file)
    
    elif FP_type == 'RDKit_FP':
        df.to_csv(dataset_file)
        featurizer = dc.feat.RDKitDescriptors()
        loader = dc.data.CSVLoader(id_field=id_field, 
                                   smiles_field=smiles_field, 
                                   tasks = task,
                                   featurizer=featurizer)
        dataset = loader.featurize(dataset_file)
        
    elif FP_type == 'Morgan': #2048
        smiles = df[smiles_field].tolist()
        mols = [Chem.MolFromSmiles(smi) for smi in smiles]
        fps = []
        for mol in mols:
            fps.append(np.matrix(AllChem.GetMorganFingerprintAsBitVect(mol,2)))
        fps = np.concatenate(fps)
        new_df = pd.DataFrame(fps)
        new_df.columns = ['V' + str(i) for i in range(fps.shape[1])]
        new_df[id_field] = df[id_field]
        for m in range(len(task)):
            new_df[task[m]] = df[task[m]]
        new_df.to_csv(dataset_file)
        user_specified_features = ['V' + str(i) for i in range(fps.shape[1])]
        featurizer = dc.feat.UserDefinedFeaturizer(user_specified_features)
        loader = dc.data.UserCSVLoader(
                tasks=task, smiles_field=smiles_field, id_field=id_field,
                featurizer=featurizer)
        dataset = loader.featurize(dataset_file)
        
    elif FP_type == 'MACCS': #167
        smiles = df[smiles_field].tolist()
        mols = [Chem.MolFromSmiles(smi) for smi in smiles]
        fps = []
        for mol in mols:
            fps.append(np.matrix(MACCSkeys.GenMACCSKeys(mol)))
        fps = np.concatenate(fps)
        new_df = pd.DataFrame(fps)
        new_df.columns = ['V' + str(i) for i in range(fps.shape[1])]
        new_df[id_field] = df[id_field]
        for m in range(len(task)):
            new_df[task[m]] = df[task[m]]
        new_df.to_csv(dataset_file)
        user_specified_features = ['V' + str(i) for i in range(fps.shape[1])]
        featurizer = dc.feat.UserDefinedFeaturizer(user_specified_features)
        loader = dc.data.UserCSVLoader(
                tasks = task, smiles_field=smiles_field, id_field=id_field,
                featurizer=featurizer)
        dataset = loader.featurize(dataset_file)
    
    else:
        print('Unsupported Fingerprint type!')
    
    return dataset, df
#----------------------------------------------
def extract_clean_dataset(subset_names, MT_df, add_features = None, id_field = 'molregno', smiles_field = 'salt_removed_smi'):
    '''
    input: subset_names --- a list of the task names
           MT_df --- pd.DataFrame of the total raw data
    output: subset of raw data， containing only the selected tasks
    '''
    extract_column = subset_names + [id_field, smiles_field]
    sub_df = MT_df[extract_column]
    
    n_tasks = len(subset_names)
    mask_mat = np.zeros((len(sub_df), n_tasks))
    for j in range(n_tasks):
        mask_mat[:,j] = np.isnan(np.array(sub_df[subset_names[j]].tolist()))
    mask_new = np.sum(mask_mat, axis = 1) < n_tasks
    
    if not add_features:
        extract_df = sub_df.loc[mask_new]
    else:
        new_sub_df = MT_df[extract_column + add_features]
        # normalization!!
        for k in range(len(add_features)):
            new_sub_df.loc[:,add_features[k]] = preprocessing.scale(np.array(new_sub_df[add_features[k]].tolist()))
        extract_df = new_sub_df.loc[mask_new]
    print('Extracted dataset shape: ' + str(extract_df.shape))
    return extract_df

#----------------------------------------------
def model_builder(model_params, model_dir):
    model = dc.models.MultitaskRegressor(**model_params)
    return model

def ST_model(n_feature, layer_size, drop_prob,
            lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0, clipnorm=5):
    training_dat = Input(shape = (n_feature,), dtype = 'float32')
    X = Dense(layer_size[0], activation = 'relu')(training_dat)
    for i in range(1,len(layer_size)):
        X = Dropout(drop_prob)(X)
        X = Dense(layer_size[i], activation = 'relu')(X) 
    model = Model(input = training_dat, output = X)
    optimizer = Adam(lr=lr, beta_1=beta_1, beta_2=beta_2, epsilon=epsilon)
    model.compile(loss = 'mean_squared_error', optimizer = optimizer)
    return model

def ST_model_hyperparam_screen(fname, task_names, FP_type, params_dict, log_path = './logs/'):
    '''
    hyperparameter screening using deepchem package
    input: fname --- name of the file of raw data containing chemicals and the value for each assay;
           task --- list of task names (supporting 1 or more, but must be a list)
           FP_type --- name of the fingprint type (one of 'Circular_2018', 'Circular_1024', 'Morgan', 'MACCS', 'RDKit_FP')
           params_dict --- dictionary containing the parameters along with the screening range
           log_path --- the directory saving the log file
    output: the log; log file saved in log_path
    '''
    log_output = []
    for task in task_names:
        print('----------------------------------------------')
        dataset_file = '%s/temp.csv' % (log_path)
        dataset = prepare_dataset(fname, [task], dataset_file, FP_type)
        for cnt in range(3):
            print('Preparing dataset for %s of rep %d...' % (task, cnt))
            splitter = dc.splits.RandomSplitter(dataset_file)
            train_dataset, valid_dataset, test_dataset = splitter.train_valid_test_split(dataset)
    
            print('Hyperprameter screening ...')
            metric = dc.metrics.Metric(dc.metrics.r2_score)
            optimizer = dc.hyper.HyperparamOpt(model_builder)
            best_dnn, best_hyperparams, all_results = optimizer.hyperparam_search(params_dict,
                                                                      train_dataset,
                                                                      valid_dataset, [],
                                                                      metric)
            # get the layer size and dropout rate of all_results
            for (key, value) in all_results.items():
                log_output.append('rep%d\t%s\t%s\t%s' % (cnt, task, str(key), str(value)))
    
            print('Generate performace report ...')
            with open('%s/hyperparam_log.txt' % (log_path), 'w') as f:
                for line in log_output:
                    f.write("%s\n" % line)
        os.system('rm %s' % dataset_file)
    
    return  log_output

def ST_model_training(fname, FP_type, best_hyperparams, result_path, epoch_num = 45,
                      frac_train = 0.8, N_test = None):
    '''
    Training the specific groups of single task (or multitask with only one task and a few chemical properties),
           using 
    input: fname --- name of the file of raw data containing chemicals and the value for each assay;
           FP_type --- name of the fingprint type (one of 'Circular_2018', 'Circular_1024', 'Morgan', 'MACCS', 'RDKit_FP')
           best_hyperparams --- dictionary containing the parameters, with key the name of the task
           result_path --- the directory saving the trained models and merits file
    output: the log; log file saved in log_path
    '''
    task_names = list(best_hyperparams.keys())
    output_dat = np.zeros((len(task_names) * 3, 12))
    rep_labels = ['NA' for _ in range(len(task_names) * 3)]
    task_labels = ['NA' for _ in range(len(task_names) * 3)]
    nrow = 0
    for task in task_names:
        print('----------------------------------------------')
        dataset_file = '%s/temp.csv' % (result_path)
        dataset = prepare_dataset(fname, [task], dataset_file, FP_type)

        for cnt in range(3):
            print('Preparing dataset for %s of rep %d...' % (task, cnt))
            splitter = dc.splits.RandomSplitter(dataset_file)
            if not N_test:
                train_dataset, test_dataset = splitter.train_test_split(dataset, frac_train = frac_train)
            else:
                frac = 1 - N_test / dataset.X.shape[0]
                train_dataset, test_dataset = splitter.train_test_split(dataset, frac_train = frac)
            
            print('Model training ...')
            n_features = train_dataset.X.shape[1]
            model = ST_model(n_features, best_hyperparams[task][0], best_hyperparams[task][1], lr=0.0001)
            for iteration in range(51):
                checkpoint_callback = ModelCheckpoint('%s/%s_rep%d_%d.hdf5' % (result_path, task,
                                                                                cnt, iteration),
                                          monitor='val_loss', verbose = 0, save_weights_only = False,
                                          mode = 'auto', period = 1)
                if iteration % 10 == 0:
                    model.fit(train_dataset.X, train_dataset.y, nb_epoch=5, callbacks = [checkpoint_callback])
                else:
                    model.fit(train_dataset.X, train_dataset.y, nb_epoch=epoch_num, verbose=0)
            
            print('Training baseline models ...')
            X_new = np.c_[[1]*train_dataset.X.shape[0], train_dataset.X]
            clf = optimize_SVR(X_new, train_dataset.y)
            reg = optimize_RidgeCV(X_new, train_dataset.y)
            
            #print('Training TRACE models ...')
            # todo!
            
            print('Saving metrics ...')
            y_true = test_dataset.y.flatten()
            y_pred_ST = model.predict(test_dataset.X).flatten()
            y_pred_SVR = clf.predict(np.c_[[1]*test_dataset.X.shape[0],test_dataset.X]).flatten()
            y_pred_Rig = reg.predict(np.c_[[1]*test_dataset.X.shape[0],test_dataset.X]).flatten()
            
            output_dat[nrow, 0] = sqrt(mean_squared_error(y_pred_ST, y_true))
            output_dat[nrow, 1] = sqrt(mean_squared_error(y_pred_SVR, y_true))
            output_dat[nrow, 2] = sqrt(mean_squared_error(y_pred_Rig, y_true))
            output_dat[nrow, 3], _ = pearsonr(y_pred_ST, y_true)
            output_dat[nrow, 4], _ = pearsonr(y_pred_SVR, y_true)
            output_dat[nrow, 5], _ = pearsonr(y_pred_Rig, y_true)
            
            y_true = train_dataset.y.flatten()
            y_pred_ST = model.predict(train_dataset.X).flatten()
            y_pred_SVR = clf.predict(np.c_[[1]*train_dataset.X.shape[0],train_dataset.X]).flatten()
            y_pred_Rig = reg.predict(np.c_[[1]*train_dataset.X.shape[0],train_dataset.X]).flatten()
            
            output_dat[nrow, 6] = sqrt(mean_squared_error(y_pred_ST, y_true))
            output_dat[nrow, 7] = sqrt(mean_squared_error(y_pred_SVR, y_true))
            output_dat[nrow, 8] = sqrt(mean_squared_error(y_pred_Rig, y_true))
            output_dat[nrow, 9], _ = pearsonr(y_pred_ST, y_true)
            output_dat[nrow, 10], _ = pearsonr(y_pred_SVR, y_true)
            output_dat[nrow, 11], _ = pearsonr(y_pred_Rig, y_true)
            
            rep_labels[nrow] = cnt
            task_labels[nrow] = task
            nrow += 1
        os.system('rm %s' % dataset_file)
    
        print('Generate performace report ...')
        output_df = pd.DataFrame(output_dat)
        output_df.columns = ['test_rmse_ST', 'test_rmse_SVR', 'test_rmse_Rig',
                             'test_pearsonr_ST', 'test_pearsonr_SVR', 'test_pearsonr_Rig',
                             'train_rmse_ST', 'train_rmse_SVR', 'train_rmse_Rig',
                             'train_pearsonr_ST', 'train_pearsonr_SVR', 'train_pearsonr_Rig']
        output_df['rep_label'] = rep_labels
        output_df['task_label'] = task_labels
        os.system('rm %s' % dataset_file)
        output_df.to_csv('%s/performance_metrics.csv' % (result_path))
    return  output_df



#!/usr/bin/env python
#-*- coding:utf-8 -*-

with open("README.md", "r") as f:
    long_description = f.read()

from setuptools import setup, find_packages

setup(
    name = "visar",
    version = "0.1.3.1",
    keywords = ("Chemoinformatics", "neural network", "visualized structure-activity relationship", "chemical landscape"),
    description = "This project aims to train neural networks by compound-protein interactions and provides interpretation of the learned model by interactively showing transformed chemical landscape and visualized SAR for chemicals of interest.",
    long_description = long_description,
    long_description_content_type='text/markdown',
    license = "MIT Licence",

    url = "https://github.com/Svvord/visar",
    author = "Qingyang Ding",
    author_email = "dingqy14@mails.tsinghua.edu.cn",
    maintainer = "S. Hou",
    maintainer_email = "housy17@mails.tsinghua.edu.cn",

    packages = find_packages(),
    include_package_data = True,
    platforms = "any",
    install_requires = [
        "numpy",
        "scipy",
        "sklearn",
        "pandas",
        "tensorflow",
        "kears",
        "align-it",
        "deepchem",
        "rdkit",
        ]
)



# -*- coding: utf-8 -*-
"""
Created on Sun Mar 10 15:34:48 2019

@author: Dingqy
"""

import numpy as np
import pandas as pd
from scipy import optimize
import math
import pickle

import deepchem as dc
from Model_training_utils import prepare_dataset

from math import sqrt
from scipy.stats import pearsonr
from sklearn.metrics import mean_squared_error

class Parameters(object):
    init_mean = 0.05
    init_sigma = 1

    def __init__(self, rown, coln):
        self._omega_matrix = self.init_sigma * \
                             np.random.randn(rown, coln) + self.init_mean
        self._omega_mean = np.zeros(coln, dtype=np.float32) + self.init_mean
        self._sigma_in_colums = np.zeros(
            coln, dtype=np.float32) + self.init_sigma
        self._sigma_in_group = np.zeros(
            rown, dtype=np.float32) + self.init_sigma

    def to_ndarray(self, rown, coln):
        result = np.zeros((rown + 2, coln + 1), np.float32)
        matrix_size = rown * coln
        result.put(range(matrix_size), self._omega_matrix)
        result.put(range(matrix_size, matrix_size + coln), self._omega_mean)
        result.put(range(matrix_size + coln, matrix_size + 2 * coln), self._sigma_in_colums)
        result.put(range(matrix_size + 2 * coln, matrix_size + 2 * coln + rown), self._sigma_in_group)
        #result.flat[range(matrix_size, matrix_size + coln)] = self._omega_mean
        #result.flat[range(matrix_size + coln, matrix_size +
        #                  2 * coln)] = self._sigma_in_colums
        #result.flat[range(matrix_size + 2 * coln, matrix_size + 2 * coln +
        #                  rown)] = self._sigma_in_group
        return result

    def from_ndarray(self, ndarray, rown, coln):
        if ndarray.size < (rown + 2) * (coln + 1):
            raise RuntimeError("ndarray size is not enough.")
        matrix_size = rown * coln
        self._omega_matrix.put(range(matrix_size), ndarray)      
        self._omega_mean = ndarray.flat[[x for x in range(matrix_size, matrix_size + coln)]]
        self._sigma_in_colums = ndarray.flat[[x for x in range(matrix_size + coln,
                                                   matrix_size + 2 * coln)]]
        self._sigma_in_group = ndarray.flat[[ x for x in range(
            matrix_size + 2 * coln, matrix_size + 2 * coln + rown)]]    
    
class TRACE(object):

    def __init__(self, rown, coln, proteins, X_train, Y_train, X_test, 
                 Y_test, df, scale_g, scale_c):
        self._rown = rown
        self._coln = coln
        self._proteins = proteins
        
        self.X_train = X_train
        self.Y_train = Y_train
        self.X_test = X_test
        self.Y_test = Y_test
        
        self.f_value = []
        
        self.df = df
        self.scale_group = scale_g
        self.scale_coln = scale_c
        
        N_train_list = []
        for i in range(rown):
            N_train_list.append(X_train[i].shape[0])
        self.N_train = N_train_list

    def loss_fun(self, ndarray):
        rown = self._rown
        coln = self._coln
        # genearte paramters of correct format
        parameter = Parameters(rown, coln)
        parameter.from_ndarray(ndarray, rown, coln)
            
        # 1. summation of likelihood function of each protein.
        tmp1 = sum(rown/2 * np.array([math.log(2 * math.pi * x) for x in parameter._sigma_in_colums]))
        omega_rep = np.matrix(parameter._omega_mean.repeat(rown, axis = 0)).reshape(coln, rown)
        delta_omega = parameter._omega_matrix - omega_rep.T
        for i in range(coln):
            tmp1 += (delta_omega[:,i].T * delta_omega[:,i]) / (2 * parameter._sigma_in_colums[i])
        # 2. summation of log format of prior distribution of sigma_y for each
        #    protein(regression epsilon's variance)
        tmp2 = sum([self.N_train[j]/2 * math.log(2* math.pi * parameter._sigma_in_group[j]) for j in range(rown)])        
        for i in range(rown):
            tmp_diff = self.Y_train[i] - np.matrix(parameter._omega_matrix[i,:]) * self.X_train[i].T 
            tmp2 += np.inner(tmp_diff, tmp_diff) / (2* parameter._sigma_in_group[i])
        # 3. summation of log format of prior distribution of omega for each
        #    protein.
        tmp3 = sum([math.log(x) / (1 + self.df/2) + self.df * self.scale_coln / (2*x) for x in parameter._sigma_in_colums])
        # 4. log format of prior distribution of variance of omega for each
        #    colum of omega matrix.
        tmp4 = sum([math.log(x) / (1 + self.df/2) + self.df * self.scale_group / (2*x) for x in parameter._sigma_in_group])
        value = float(tmp1 + tmp2 + tmp3 + tmp4)
        self.f_value.append(value)
        return value
    
    def fprime(self, p):
        rown = self._rown
        coln = self._coln
        # genearte paramters of correct format
        parameter = Parameters(rown, coln)
        parameter.from_ndarray(p, rown, coln)
        output = np.zeros(p.shape)
        matrix_size = coln * rown
        
        #1 fprime of omega_matrix
        ## calculate row by row (iterate over each protein of the family)
        for i in range(rown):
            term1 = (parameter._omega_matrix[i,:] - parameter._omega_mean) / parameter._sigma_in_colums
            tmp_diff = self.Y_train[i] - np.matrix(parameter._omega_matrix[i,:]) * self.X_train[i].T
            term = term1 - 2 * tmp_diff * self.X_train[i]
            output.flat[[x for x in range(i * coln, (i+1) * coln)]] = term
        
        #2 fprime of omega_mean
        omega_rep = np.matrix(parameter._omega_mean.repeat(rown, axis = 0)).reshape(coln, rown)
        delta_omega = parameter._omega_matrix - omega_rep.T
        output.flat[[x for x in range(matrix_size, matrix_size + coln)]] = - np.sum(delta_omega, axis = 0) / parameter._sigma_in_colums
        
        #3 fprime of sigma in column
        term3 = (1 / (1+ self.df/2) + rown / 2) * (1 / parameter._sigma_in_colums)
        term4 = np.array([(np.inner(delta_omega[:,j].T, delta_omega[:,j].T) + self.df*self.scale_coln) /(2 * parameter._sigma_in_colums[j]**2) for j in range(coln)]).flatten()
        output.flat[[x for x in range(matrix_size + coln, matrix_size + 2* coln)]] = term3 - term4
        
        #4 fprime of sigma in group
        term5 = (1/(1+self.df/2) + np.array(self.N_train) / 2) * (1 / parameter._sigma_in_group)
        for j in range(rown):
            tmp_diff = self.Y_train[j] - np.matrix(parameter._omega_matrix[j,:]) * self.X_train[j].T
            term6 = (np.array(np.inner(tmp_diff, tmp_diff)  + self.df*self.scale_group) / (2 * parameter._sigma_in_group[j]**2))[0]
            output.flat[matrix_size + 2* coln + j] = term5[j] - term6
        return output
    
    def set_bound(self):
        # * (min, max) for each x
        rown = self._rown
        coln = self._coln
        
        matrix_size = rown * coln
        bnds = [(None, None) for i in range(matrix_size + coln)] + \
           [(0.01, None) for i in range(matrix_size + coln, matrix_size + 2 * coln + rown)] + \
           [(0,0), (0,0)]
        bnds = tuple(x for x in bnds)
        return bnds

def fmin_trace(target, MAX_ITERATION = 15000):
    OPT_METHOD = "L-BFGS-B"
    #MAX_ITERATION = 15000
    
    init_point = Parameters(target._rown, target._coln).to_ndarray(target._rown, target._coln)
    res = optimize.minimize(target.loss_fun, init_point, method = OPT_METHOD,
                            jac = target.fprime,
                            hess=False,
                            bounds=target.set_bound(),
                            options={'maxiter': MAX_ITERATION,
                                     'disp': True,
                                     'maxfun': 15000})
    return res, target.f_value


def TRACE_predict(Xj, trace_obj, param_values, j):
    rown = trace_obj._rown
    coln = trace_obj._coln
    parameter_result = Parameters(rown, coln)
    parameter_result.from_ndarray(param_values, rown, coln)
    omega_matrix = parameter_result._omega_matrix
    y_pred = np.array(np.matrix(omega_matrix[j,:]) * Xj.T)[0]
    return y_pred

def prepare_TRACE_dataset(fname, task_names, FP_type, result_path):
    X_train = []
    X_test = []
    X_val = []
    Y_train = []
    Y_test = []
    Y_val = []
    for task in task_names:
        print('----- Processing dataset of %s -------' % task)
        dataset_file = '%s/temp.csv' % (result_path)
        dataset = prepare_dataset(fname, [task], dataset_file, FP_type)
        
        for rep in range(3):
            splitter = dc.splits.RandomSplitter(dataset_file)
            train_dataset, valid_dataset, test_dataset = splitter.train_valid_test_split(dataset)
        
            X_train.append(np.c_[[1]*train_dataset.X.shape[0], train_dataset.X])
            X_test.append(np.c_[[1]*test_dataset.X.shape[0], test_dataset.X])
            X_val.append(np.c_[[1]*valid_dataset.X.shape[0], valid_dataset.X])
            Y_train.append(train_dataset.y.flatten())
            Y_test.append(test_dataset.y.flatten())
            Y_val.append(valid_dataset.y.flatten())
    
    for rep in range(3):
        X_train0 = [X_train[k] for k in range(len(X_train)) if k % 3 == rep]
        X_test0 = [X_test[k] for k in range(len(X_train)) if k % 3 == rep]
        X_val0 = [X_val[k] for k in range(len(X_train)) if k % 3 == rep]
        Y_train0 = [Y_train[k] for k in range(len(X_train)) if k % 3 == rep]
        Y_test0 = [Y_test[k] for k in range(len(X_train)) if k % 3 == rep]
        Y_val0 = [Y_val[k] for k in range(len(X_train)) if k % 3 == rep]
        pickle.dump([X_train0, X_test0, X_val0, Y_train0, Y_test0, Y_val0], 
                    open(result_path + '/temp_%d_dataset.p' % (rep), 'wb'))
    return

def TRACE_model_hyperparam_screen(fname, task_names, FP_type, hyperparams, 
                                  MAX_ITERATION = 100, log_path = './logs/'):
    log_output = []
    output_store = []
    hyperparam_record = []
    for k in range(len(hyperparams)):
        for rep in range(3):
            print('---- processing hyperparam %s of rep %d ----' % (str(hyperparams[k]), rep))
            [X_train,_,X_val,Y_train,_,Y_val]=pickle.load(open(log_path+'/temp_%d_dataset.p'%(rep),'rb'))
            rown = len(X_train)
            coln = X_train[0].shape[1]
            output_dat = np.zeros((rown,2))
            df = hyperparams[k][0]
            scale_g = hyperparams[k][1]
            scale_c = hyperparams[k][2]
            trace_obj = TRACE(rown, coln, task_names, X_train, Y_train, X_val, Y_val, 
                  df, scale_g, scale_c)
            res, values = fmin_trace(trace_obj, MAX_ITERATION)
            for j in range(rown):
                y_pred = TRACE_predict(X_val[j], trace_obj, res['x'], j)
                y_real = Y_val[j]
                output_dat[j,0] = sqrt(mean_squared_error(y_pred, y_real))
                output_dat[j,1] = pearsonr(y_pred, y_real)[0]
            output_store.append(np.array(output_dat))
        log_output.append(np.concatenate(output_store, axis = 1))
        output_store = []
        hyperparam_record = hyperparam_record + [str(hyperparams[k]) for _ in range(rown)]
        out = np.concatenate(log_output, axis = 0)
        df = pd.DataFrame(out)
        df['label'] = hyperparam_record
        df.to_csv(log_path + 'hyperparam_screen.csv')
    return df

def TRACE_model_training(task_names, log_path, params, output_prefix):
    task_label = ['' for _ in range(len(task_names) * 3)]
    rep_label = ['' for _ in range(len(task_names) * 3)]
    output_dat = np.zeros((len(task_names) * 3 ,4))
    cnt = 0
    for rep in range(3):
        print('---- processing rep %d ----' % (rep))
        [X_train,_,X_test,Y_train,_,Y_test]=pickle.load(open(log_path+'/temp_%d_dataset.p'%(rep),'rb'))
        rown = len(X_train)
        coln = X_train[0].shape[1]
        df = params[0]
        scale_g = params[1]
        scale_c = params[2]
        trace_obj = TRACE(rown, coln, task_names, X_train, Y_train, X_test, Y_test, 
              df, scale_g, scale_c)
        res, values = fmin_trace(trace_obj)
        for j in range(rown):
            y_pred = TRACE_predict(X_train[j], trace_obj, res['x'], j)
            y_real = Y_train[j]
            output_dat[cnt,0] = sqrt(mean_squared_error(y_pred, y_real))
            output_dat[cnt,1] = pearsonr(y_pred, y_real)[0]
            
            y_pred = TRACE_predict(X_test[j], trace_obj, res['x'], j)
            y_real = Y_test[j]
            output_dat[cnt,2] = sqrt(mean_squared_error(y_pred, y_real))
            output_dat[cnt,3] = pearsonr(y_pred, y_real)[0]
            task_label[cnt] = task_names[j]
            rep_label[cnt] = 'rep' + str(rep)
            cnt += 1
            
        df = pd.DataFrame(output_dat)
        df['task'] = task_label
        df['rep'] = rep_label
        df.to_csv(log_path + output_prefix + '.csv')
        
        pickle.dump(res['x'], open(output_prefix + '_' + str(rep) + '.p', 'wb'))

